@@ -5,7 +5,7 @@ kombu.transport.virtual.exchange
 Implementations of the standard exchanges defined
 by the AMQ protocol  (excluding the `headers` exchange).
 
-:copyright: (c) 2009 - 2010 by Ask Solem.
+:copyright: (c) 2009 - 2011 by Ask Solem.
 :license: BSD, see LICENSE for more details.
 
 """
@@ -18,6 +18,7 @@ class ExchangeType(object):
     :param channel: AMQ Channel
 
     """
+    type = None
 
     def __init__(self, channel):
         self.channel = channel
@@ -46,15 +47,23 @@ class ExchangeType(object):
 
 class DirectExchange(ExchangeType):
     """The `direct` exchange routes based on exact routing keys."""
+    type = "direct"
 
     def lookup(self, table, exchange, routing_key, default):
         return [queue for rkey, _, queue in table
                     if rkey == routing_key] or [default]
 
+    def deliver(self, message, exchange, routing_key, **kwargs):
+        _lookup = self.channel._lookup
+        _put = self.channel._put
+        for queue in _lookup(exchange, routing_key):
+            _put(queue, message, **kwargs)
+
 
 class TopicExchange(ExchangeType):
     """The `topic` exchanges routes based on words separated by dots, and
     wildcard characters `*` (any single word), and `#` (one or more words)."""
+    type = "topic"
 
     #: map of wildcard to regex conversions
     wildcards = {"*": r".*?[^\.]",
@@ -64,8 +73,16 @@ class TopicExchange(ExchangeType):
     _compiled = {}
 
     def lookup(self, table, exchange, routing_key, default):
-        return  [queue for rkey, pattern, queue in table
-                            if self._match(pattern, routing_key)] or [default]
+        return [queue for rkey, pattern, queue in table
+                        if self._match(pattern, routing_key)] or [default]
+
+    def deliver(self, message, exchange, routing_key, **kwargs):
+        _lookup = self.channel._lookup
+        _put = self.channel._put
+        deadletter = self.channel.deadletter_queue
+        for queue in [q for q in _lookup(exchange, routing_key)
+                            if q and q != deadletter]:
+            _put(queue, message, **kwargs)
 
     def prepare_bind(self, queue, exchange, routing_key, arguments):
         return routing_key, self.key_to_pattern(routing_key), queue
@@ -96,10 +113,15 @@ class FanoutExchange(ExchangeType):
     for an example implementation of these methods.
 
     """
+    type = "fanout"
 
     def lookup(self, table, exchange, routing_key, default):
         return [queue for _, _, queue in table]
 
+    def deliver(self, message, exchange, routing_key, **kwargs):
+        if self.channel.supports_fanout:
+            self.channel._put_fanout(exchange, message, **kwargs)
+
 
 #: Map of standard exchange types and corresponding classes.
 STANDARD_EXCHANGE_TYPES = {"direct": DirectExchange,
